// Description: detects a wand pointing at any geometry below it and produces an event
//
//<b>notes:</b>
//<ul>
//<li> derived nodes inherit all messages from their base classes
//<li> see reset method for default settings
//<li> all wand events are also generated by this node (ex. button1)
//<li> the local mode determines if the intersection position is local to the object or relative to world coordinates
//</ul>
//
// Category: Trigger
// Author: Dave Pape
// Revision: 11/01/01
//
#include <Performer/pf/pfGroup.h>
#include <Performer/pr/pfGeoSet.h>
#include <Performer/pf/pfGeode.h>
#include <ygNodeDB.h>
#include <ygWorld.h>
#include <ygUser.h>
#include <ygWand.h>
#include <ygNetMessages.h>
#include "pointNodeTrigger.h"

extern "C" ygNode* construct_pointNodeTrigger(const char* name,bool master) { return new pointNodeTrigger(name,master); }

pointNodeTrigger::pointNodeTrigger(const char* name,bool master) : ygNode(name,master)
    {
    setClassName("pointNodeTrigger");
    debugFlag("pointing", &debugPointing);
    detectType = "ygNode";
    distance = 10;
    isectType = YG_ISECT_ALLMASK;
    addNetMessage("detect",&detectType,YG_NET_CLASS);
    addNetMessage("distance",&distance,YG_NET_FLOAT);
    addNetMessage("intersect",&isectType,YG_NET_INT);
    addNetMessageMap("intersect","all",YG_ISECT_ALLMASK);
    addNetMessageMap("intersect","wall",YG_ISECT_WALLMASK);
    addNetMessageMap("intersect","floor",YG_ISECT_FLOORMASK);
    addNetMessageMap("intersect","grab",YG_ISECT_GRABMASK);
    addNetMessageMap("intersect","debug",YG_ISECT_DEBUGMASK);
    addNetMessageMap("intersect","off",0);
    addEvent("start");
    addEvent("stop");
    }

pointNodeTrigger::~pointNodeTrigger(void)
    {
    }


void pointNodeTrigger::reset(void)
    {
    prevpointing.clear();
    //set detect type to ygNode
    detectType = "ygNode";
    //set distance to 10.0
    distance = 10;
    //set intersect class to "all"
    isectType = YG_ISECT_ALLMASK;
    //for each user in the world
    for (int i=0; i < ygWorld::world()->numUsers(); i++)
        {
        ygUser* user = ygWorld::world()->user(i);
        //for each associated wand
        for (int j=0; j < user->numWands(); j++)
            {
            if (user->wand(j) && user->wand(j)->isOfClass("ygWand"))
                {
                ygWand* wand = (ygWand*) (user->wand(j));
                const vector<ygString>& events = getEvents();
                const vector<ygString>& wandEvents = wand->getEvents();
                //for each wand event
                for (int k=0; k<wandEvents.size(); k++)
                    {
                    bool found = false;
                    for (int l=0; l<events.size(); l++)
                        {
                        if (events[l] == wandEvents[k])
                            found = true;
                        }
                    //if not found then add event
                    if (!found)
                        addEvent(wandEvents[k]);
                    }
                }
            }
        }
    ygNode::reset();
    }


void pointNodeTrigger::message(const ygMessage& msg)
    {
    //set the type of node to detect
    if (msg == "detect")
        {
        return;
        }
    //set the intersection distance from the wand
    else if (msg == "distance")
        {
        return;
        }
    //set the intersection class to detect
    else if (msg == "intersect")
        {
        return;
        }
    else
        ygNode::message(msg);
    }


void pointNodeTrigger::app(void)
    {
    set<ygWand*> pointing;
    int i,j,f;
    pfHit **hits[32];
    pfMatrix myXform, myXformInv;
    getTransform(myXform);
    myXformInv.invertAff(myXform);
    //for all users in the world
    for (i=0; i < ygWorld::world()->numUsers(); i++)
        {
        //for each user wand
        ygUser * user = ygWorld::world()->user(i);
        for (j=0; j < user->numWands(); j++)
            {
            if (user->wand(j) && user->wand(j)->isOfClass("ygWand"))
                {
                ygWand * wand = (ygWand*) (user->wand(j));
                //if the wand is pointing at geometry
                if ((wand) && (pointingAtMe(wand,myXformInv,hits)))
                    {
                    pfGeode* geode = NULL;
                    (*hits[0])->query(PFQHIT_NODE, &geode);
                    if (geode)
                        {
                        ygNode* node=NULL;
			//if (geode->getName())
                        	node = ygNodeDB::find(geode->getName(),false);
                        if (node && node->isOfClass(detectType))
                            {
                            pointing.insert(pointing.begin(),wand);
                            pfMatrix mat;
                            (*hits[0])->query(PFQHIT_XFORM, (float*)mat.mat);
                            if (debugPointing)
                                cout << name() << ": " << user->name() << " is pointing\n";
                            ygString args("wand=");
                            args += wand->name();
                            args += " user=";
                            args += user->name();
                            args += " node=";
                            args += geode->getName();
                            char value[32];
                            args += " xsize=";
                            sprintf(value,"%f",mat.mat[0][0]);
                            args += value;
                            args += " ysize=";
                            sprintf(value,"%f",mat.mat[1][1]);
                            args += value;
                            args += " zsize=";
                            sprintf(value,"%f",mat.mat[2][2]);
                            args += value;
                            mat.mat[0][0] = 1.0;
                            mat.mat[1][1] = 1.0;
                            mat.mat[2][2] = 1.0;
                            pfCoord coord;
                            mat.getOrthoCoord(&coord);
                            args += " xpos=";
                            sprintf(value,"%f",coord.xyz[0]);
                            args += value;
                            args += " ypos=";
                            sprintf(value,"%f",coord.xyz[1]);
                            args += value;
                            args += " zpos=";
                            sprintf(value,"%f",coord.xyz[2]);
                            args += value;
                            args += " xori=";
                            sprintf(value,"%f",coord.hpr[1]);
                            args += value;
                            args += " yori=";
                            sprintf(value,"%f",coord.hpr[2]);
                            args += value;
                            args += " zori=";
                            sprintf(value,"%f",coord.hpr[0]);
                            args += value;
                            //printf("args: %s\n",args.c_str());
                            //if the number of wand flags is positive
                            if (wand->numFlags() > 0)
                                {
                                ygString event;
                                //for each wand flag
                                for (f=0; f < wand->numFlags(); f++)
                                    {
                                    wand->getFlag(f,event);
                                    //any wand events are generated while in the space
                                    eventOccurred(event,args);
                                    }
                                }
                            //if not previously pointing generate start event
                            if (prevpointing.find(wand) == prevpointing.end())
                                {
                                //a wand and associated user have begun pointing
                                eventOccurred("start",args);
                                }
                            }
                        }
                    }
                }
            }
        }
    set<ygWand*>::const_iterator iter;
    //for each wand previously pointing at geometry
    for (iter = prevpointing.begin(); iter != prevpointing.end();
         ++iter)
        //if not currently pointing then generate stop event
        if (pointing.find(*iter) == pointing.end())
            {
            ygString args("wand=");
            args += (*iter)->name();
            args += " user=";
            args += (*iter)->user()->name();
            //a wand and associated user have stopped pointing
            eventOccurred("stop",args);
            }
    prevpointing = pointing;
    ygNode::app();
    }


bool pointNodeTrigger::pointingAtMe(const ygWand* wand,const pfMatrix& myXformInv,pfHit*** hits) const
    {
    if (isectType)
        {
        pfSegSet segset;
        pfMatrix wandXform;
        pfVec3 forward(0, 1, 0), wandDir;
        segset.activeMask = 1;
        segset.isectMask = (uint)isectType;
        segset.discFunc = NULL;
        segset.bound = NULL;
        segset.mode = PFTRAV_IS_PRIM;
        segset.segs[0].pos = wand->origin(this);
        wand->getTransform(wandXform);
        wandDir.xformVec(forward, wandXform);
        wandDir.xformVec(wandDir, myXformInv);
        segset.segs[0].dir = wandDir;
        segset.segs[0].length = distance;
        return pfnode()->isect(&segset, hits);
        }
    return false;
    }
