// Description: generates events describing node names
//
//<b>notes:</b>
//<ul>
//<li> derived nodes inherit all messages from their base classes
//<li> see reset method for default settings
//</ul>
//
// Category: System
// Author: Alex Hill
// Revision: 09/15/05 Alex Hill
//
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <ygNodeDB.h>
#include <ygNetMessages.h>
#include <ygDSOLoader.h>
#include "nodeListing.h"

using namespace std;

extern "C" ygNode* construct_nodeListing(const char* name,bool master) { return new nodeListing(name,master); }

nodeListing::nodeListing(const char* name,bool master) : ygNode(name,master)
    {
    setClassName("nodeListing");
    index = 0;
    maximum = 100;
    node = NULL;
    addNetMessage("node",&nodeName,YG_NET_CLASS);
    addNetMessage("class",&className,YG_NET_CLASS);
    addNetMessage("class",&index,YG_NET_INT);
    addNetMessage("maximum",&maximum,YG_NET_INT);
    addEvent("class");
    addEvent("changed");
    }
    
void nodeListing::reset(void)
    {
    node = NULL;
    lastIndex = -1;
    //clear the node name
    resetNetMessage("node");
    //clear the class name
    resetNetMessage("class");
    //set the maximum printout to 100
    resetNetMessage("maximum");
    nodes.clear();
    ygNode::reset();
    }

void nodeListing::message(const ygMessage& msg)
    {
    //find node names below the given node
    if (msg == "node")
        {
        if (nodeName != lastNode)
            {
            if (nodeName.length() > 0)
                node = ygNodeDB::find(nodeName,false);
            else
                node = NULL;
            lastNode = nodeName;
            }
        checkClass();
        }
    //generate events with node names from the given class and index
    else if (msg == "class")
        {
        checkClass();
        }
    //generate events with class names using the given index
    else if (msg == "index")
        {
        if (msg.numArgs() > 0)
            {
            index = msg.intArg(0);
            lastIndex = -1;
            checkClass();
            }
        else
            msg.error(name(),"(wrong number of arguments)");
        }
    //set the total number of classes to report
    else if (msg == "maximum")
        {
        return;
        }
    else
        ygNode::message(msg);
    }


void nodeListing::findNodes(ygNode* currNode)
    {
    if (currNode->isOfClass(className))
        {
        ygString currName;
        currName = currNode->name();
        vector<ygString>::iterator i;
        //order currClasses in alphabetical order
        for (i=nodes.begin();i!=nodes.end();i++)
            {
            if (currName < (*i))
                break;
            }
        //insert the new directory into the list
        nodes.insert(i,currName);
        }
    //call findClass recursively on all children
    for (int i=0; i < currNode->numChildren(); i++)
        findNodes(currNode->child(i));
    }


void nodeListing::app(void)
    {
    if (node)
        {
        if (!changed)
            {
            if (node->isChanged(YG_CHANGE_TREE,false))
                {           
                //populate current node list
                nodes.clear();
                findNodes(node);
                changed = true;
                lastIndex = -1;
                checkClass();
                }
            }
        else
            {
            node->isChanged(YG_CHANGE_TREE);
            changed = false;
            }
        }
    else
        changed = false;
    ygNode::app();
    }


void nodeListing::checkClass(void)
    {
    if (className != lastClass)
        {
        if (node)
            {
            nodes.clear();
            findNodes(node);
            lastIndex = -1;
            }
        lastClass = className;
        }
    if (index != lastIndex)
        {
        int nodeCount;
        nodeCount = nodes.size();
        int nodeMax = index + maximum;
        if (nodeMax > nodeCount)
            nodeMax = nodeCount;
        ygString args("total=");
        char val[32];
        sprintf(val,"%d",nodeCount);
        args += val;
        eventOccurred("changed",args);
        int num=0;
        for (int i=index;i<nodeMax;i++)
            {
            ygString args("name=");
            args += nodes[i];
            args += " number=";
            sprintf(val,"%d",num);
            args += val;
            eventOccurred("class",args);
            num++;
            }
        }
    lastIndex = index;
    }
