// Description: a navigator that takes a direction and rotation vector
//
//<b>notes:</b>
//<ul>
//<li> all parent node messages are inherited
//<li> see reset method for default settings
//<li> a direction of (0,1,0) will move in the direction of the front wall
//</ul>
//
// Category: User
// Author: Alex Hill
// Revision: 12/05/03
//
#include <ygWorld.h>
#include <ygNetMessages.h>
#include <ygNodeDB.h>
#include "vectorNavigator.h"

extern "C" ygNode* construct_vectorNavigator(const char* name,bool master) { return new vectorNavigator(name,master); }

vectorNavigator::vectorNavigator(const char* name,bool master) : ygCAVENavigator(name,master)
    {
    setClassName("vectorNavigator");
    direction.set(0,0,0);
    rotation.set(0,0,0);
    addNetMessage("direction",&direction,YG_NET_VEC3);
    addNetMessage("rotation",&rotation,YG_NET_VEC3);
    }

vectorNavigator::~vectorNavigator(void)
    {
    }

void vectorNavigator::reset(void)
    {
    //set directon vector to 0,0,0
    direction.set(0,0,0);
    //set rotation vector to 0,0,0
    rotation.set(0,0,0);
    ygCAVENavigator::reset();
    }

void vectorNavigator::message(const ygMessage& msg)
    {
    //set the direction vector
    if (msg == "direction")
        {
        return;
        }
    //set the rotation vector
    else if (msg == "rotation")
        {
        return;
        }
    //set the absolute orientation
    else if (msg == "orientation")
        {
        if (msg.numArgs() == 3)
            {
            pfVec3 ori;
            msg.getVec3Args(ori);
            setOrientation(ori);
            }
        else
            msg.error(name(),"(wrong number of arguments)");
        }
    else
        ygCAVENavigator::message(msg);
    }

void vectorNavigator::app(void)
    {
    //process direction and orientation input
    checkInput();
    //if fly mode is false then adjust height to any ground collision
    if (!fly_)
        checkGround();
    //if collision detection is on then test for object collisions
    if (collide_)
        checkCollision();
    ygNavigator::app();
    }


void vectorNavigator::checkInput(void)
    {
    //extract X and Z vector components
    pfVec3 rot,ori;
    ori = orientation();
    rot[0] = rotation[2];
    rot[1] = 0.0;
    rot[2] = -rotation[0];
    pfMatrix mat;
    mat.makeEuler(ori[2],ori[0],ori[1]);
    pfVec3 currDir(0,1,0),newDir(0,1,0);
    currDir.xformVec(currDir,mat);
    //add X and Z components to forward vector (0,1,0)
    newDir += rot * (rotSpeed_/90.0) * ygWorld::frameDeltaTime();
    newDir.normalize();
    //transform direction vector by current navigation matrix and normalize
    newDir.xformVec(newDir,mat);
    pfMatrix headingPitch;
    headingPitch.makeVecRotVec(currDir,newDir);
    pfMatrix newMat;
    //multiply existing navigation matrix by new heading and pitch matrix
    newMat = mat * headingPitch;
    float newRoll;
    newRoll = rotation[1] * rotSpeed_ * ygWorld::frameDeltaTime();
    pfMatrix roll;
    //use Y vector component to create rotation matrix
    roll.makeRot(newRoll,0,1,0);
    //multiply existing navigation matrix by new roll matrix
    newMat *= roll;
    //calculate translation component
    pfVec3 dir, pos;
    pos = position();
    dir = direction;
    //if fly mode is false then zero the Z component
    if (!fly_)
        dir[PF_Z] = 0;
    //transform the direction vector by the user matrix
    dir.xformPt(dir,mat);
    //multiply existing navigation matrix by new translation matrix
    pos += dir * transSpeed_ * ygWorld::frameDeltaTime();
    newMat.postTrans(newMat,pos[0],pos[1],pos[2]);
    //update navigation coordinates
    pfCoord coord;
    newMat.getOrthoCoord(&coord);
    setOrientation(coord.hpr[1],coord.hpr[2],coord.hpr[0]);
    setPosition(coord.xyz[0],coord.xyz[1],coord.xyz[2]);
    }






















