// Author: Dave Pape
// Revision: 06/28/05 Alex Hill - add functionality to replace node names in messages
// Revision: 02/01/06 Alex Hill - add functionality to handle frame count delays          
//
#include <stdlib.h>
#include "ygNodeDB.h"
#include "ygEvent.h"
#include "ygNode.h"
#include "ygMessage.h"

using namespace std;


ygMessage::ygMessage(void)
    {
    delay_ = 0;
    recipient_ = NULL;
    }


ygMessage::ygMessage(const ygString& name)
    {
    delay_ = 0;
    recipient_ = NULL;
    message_ = name;
    }


ygMessage& ygMessage::operator=(const ygMessage& other)
    {
    message_ = other.message_;
    args_ = other.args_;
    recipient_ = other.recipient_;
    recipientName_ = other.recipientName_;
    delay_ = other.delay_;
    return *this;
    }


ygMessage::~ygMessage(void)
    {
    }


ostream& operator<<(ostream& s,const ygMessage& m)
    {
    if (m.recipientName_.length() > 0)
        s << m.recipientName_ << ".";
    s << m.message_;
    if (m.args_.size() > 0)
        {
        int i;
        s << "(";
        for (i=0; i < m.args_.size()-1; i++)
            s << m.args_[i] << ",";
        s << m.args_[i] << ")";
        }
    if (m.delay_ > 0)
        s << "+" << m.delay_;
    else if (m.delay_ < 0)
        {
        s << "+";
        if (m.delay_ == -2.0)
            s << "+";
        else if (m.delay_ < -2.0)
            s << (int)(-m.delay_) << "f";
        }
    return s;
    }


void ygMessage::writeString(ygString& str,bool all) const
    {
    str.clear();
    if (all && recipientName_.length() > 0)
        {
        str += recipientName_;
        str += ".";
        }
    str += message_;
    if (args_.size() > 0)
        {
        int i;
        str += "(";
        for (i=0; i < args_.size()-1; i++)
            {
            str += "\"";
            str += args_[i];
            str += "\",";
            }
        str += args_[i];
        str += ")";
        }
    if (all)
        {
        if (delay_ > 0)
            {
            char delayStr[32];
            sprintf(delayStr,"+%.4f",delay_);
            str += delayStr;
            }
        else if (delay_ < 0)
            {
            char delayStr[32];
            if (delay_ < -2.0)
                sprintf(delayStr,"+%.4ff",-delay_);
            if (delay_ == -2.0)
                sprintf(delayStr,"++");
            else
                sprintf(delayStr,"+");
            str += delayStr;
            }
        }
    }


#define SKIPWHITESPACE(s,i)  \
        { if (i<s.length()) i = s.find_first_not_of(" \t",i); }

/* Parse a string, extracting individual pieces of message data
   (i.e. the recipient name, message name, argument list, and delay)
   The basic format of a message is:
    node.message(arg1, ..., argN)+delay
   where everything except 'message' is optional.  Arguments may
   be separated by whitespace or commas.
   Quotes may be used to enclose arguments which contain spaces or commas
   For example:
    foo
    foo()
    foo+1.5
    a.foo (1,2)        - args are '1' and '2'
    foo( 1 2 3) +7        - args are '1', '2', and '3'
    a.foo("1 2" 3)        - args are '1 2' and '3'
    foo ( b.bar( "1 2") 7 )    - args are 'b.bar( "1 2")' and '7'
*/
void ygMessage::parseString(const ygString& origstr,ygNode *n,ygEvent* event)
    {
    ygString str(origstr);
    ygString tmpmsg;
    int i=0, next;
    if (event)
        replaceArgs(str,n,event);
    delay_ = 0;
    SKIPWHITESPACE(str,i)
    next = str.find_first_of("()+, \t",i);
    tmpmsg.assign(str,i,next-i);
    i = next;
    splitMessageName(tmpmsg,n);
    SKIPWHITESPACE(str,i)
    if (str[i] == '(')
        {
        i++;
        SKIPWHITESPACE(str,i)
        while ((i < str.length()) && (str[i] != ')'))
            {
            ygString arg;
            if (str[i] == '"')
                {
                i++;
                while (str[i] != '"')
                    {
                    if (str[i] == '\\')
                        i++;
                    if (i < str.length())
                        {
                        arg += str[i];
                        i++;
                        }
                    else
                        break;
                    }
                if ((i < str.length()) && (str[i] == '"'))
                    i++;
                SKIPWHITESPACE(str,i)
                }
            else
                {
                int parenDepth=0;
                while ((i < str.length()) && ((parenDepth) ||
                    (!strchr("), \t\"",str[i]))))
                    {
                    if (str[i] == '\\')
                        i++;
                    else if (str[i] == '(')
                        parenDepth++;
                    else if (str[i] == ')')
                        parenDepth--;
                    if (i < str.length())
                        {
                        arg += str[i];
                        i++;
                        }
                    }
                if (parenDepth)
                    cerr << "WARNING: unclosed parenthesis in message '"
                        << str << "'\n";
                SKIPWHITESPACE(str,i)
                }
            if ((i < str.length()) && (str[i] == ','))
                i++;
            /*  Remove any trailing whitespace from the argument */
            int tail = arg.find_last_not_of(" \t");
            if (tail < arg.length()-1)
                arg.erase(tail+1);
            args_.push_back(arg);
            SKIPWHITESPACE(str,i)
            }
        if ((i < str.length()) && (str[i] == ')'))
            i++;
        }
    SKIPWHITESPACE(str,i)
    if ((i < str.length()) && (str[i] == '+'))
        {
        if (i == str.length()-1)
            delay_ = -1.0;
        else if (i == str.length()-2 && str[str.length()-1] == '+')
            delay_ = -2.0;
        else if (str[str.length()-1] == 'f')
            {
            ygString tmp;
            tmp.assign(str,i+1,(str.length()-1)-(i+1));
            delay_ = -atof(tmp.c_str());
            }
        else
            delay_ = atof(str.c_str()+i+1);
        }
    }


void ygMessage::splitMessageName(const ygString& m,ygNode *n)
    {
    int p = m.find('.');
    if ((p != m.npos) && (p > 0))
        {
        recipientName_.assign(m,0,p);
        message_.assign(m,p+1,m.length()-p-1);
        recipient_ = ygNodeDB::find(recipientName_,false);
        }
    else
        {
        recipient_ = n;
        message_ = m;
        }
    }


void ygMessage::replaceArgs(ygString& str,ygNode* node,ygEvent* event)
    {
    int dollar = str.find('$');
    while (dollar != str.npos)
        {
        if (dollar == str.length()-1)
            break;
        else if (str[dollar+1] == '$')
            str.replace(dollar,2,"$");
        else
            {
            int end = str.find_first_of(" \t\".,+(){}[]",dollar);
            ygString arg;
            arg.assign(str,dollar+1,end-dollar-1);
            ygString* result;
            if (event && (result = event->argValue(arg)))
                {
                str.replace(dollar,end-dollar,*(result));
                if (result->length() > 0)
                    dollar += result->length()-1;
                }
            else if (node && (result = node->getNetMessageArg(arg)))
                {
                str.replace(dollar,end-dollar,*(result));
                if (result->length() > 0)
                    dollar += result->length()-1;
                delete result;
                }
            else
                str.replace(dollar,end-dollar,"");
            }
        dollar = str.find('$',dollar+1);
        }
    }


void ygMessage::replaceNodeName(const ygString& from,const ygString& to)
    {
    if (recipientName_ == from)
        {
        recipientName_ = to;
        recipient_ = ygNodeDB::find(to,false);
        }
    for (int i=0;i < args_.size(); i++)
        {
        ygString* token;
        int j=0;
        while (token = args_[i].nextToken(" \t\".,=(){}",&j))
            {
            if (*token == from)
               {
               if (j == ygString::npos)
                   j = args_[i].size();
               args_[i].replace(j-from.length(),from.length(),to);
               j = 0;
               }
            delete token;
            }
        }
    }


void ygMessage::error(const ygString& prefix,const ygString& suffix) const
    {
    cerr << "ERROR: " << prefix << " received bad message: "
        << message_ << " " << suffix << endl;
    }
