// Description: a node that loads a scene file
//
//<b>notes:</b>
//<ul>
//<li> derived nodes inherit all messages from their base classes
//<li> see reset method for default settings
//<li> turing of C preprocessing can speed up long load times
//<li> note: this node does not recreate scene files upon reset
//</ul>
//
// Category: Geometry
// Author: Alex Hill
//           09/01/04
// Revision: 02/01/06 Alex Hill - added "updated" event
//
#include <ygWorld.h>
#include <ygNetMessages.h>
#include <ygParser.h>
#include <ygMessage.h>
#include "ygScene.h"

using namespace std;

extern "C" ygNode* construct_ygScene(const char* name,bool master) { return new ygScene(name,master); }


ygScene::ygScene(const char* name,bool master) : ygNode(name,master)
    {
    setClassName("ygScene");
    preprocess = true;
    addNetMessage("file",&filename,YG_NET_FILE);
    addNetMessage("file",&preprocess,YG_NET_BOOL);
    loaded = false;
    first = true;
    }


ygScene::~ygScene(void)
    {
    }


void ygScene::reset(void)
    {
    //set the preprocess flag to true
    clearEventMessages();
    const vector<ygMessage*>& messages = getInitMessages();
    vector<ygMessage*>::const_iterator iter;
    for (iter=messages.begin(); iter != messages.end(); ++iter)
        {
        if (loaded || (*iter)->recipient() == this)
            ygWorld::world()->scheduleMessage(**iter,true);
        }
    }


void ygScene::message(const ygMessage& msg)
    {
    //name of scene file to load with optional preprocess flag
    if (msg == "file")
        {
        loadFile(filename,preprocess);
        }
    //delete scene related nodes and reload scene file
    else if (msg == "reload")
        {
        set<ygNode*> allChildren;
        for (int i=0;i<numChildren();i++)
            allChildren.insert(allChildren.begin(),child(i));
        set<ygNode*>::const_iterator iter;
        for (iter = sceneChildren.begin(); iter != sceneChildren.end();++iter)
            {
            if (allChildren.find(*iter) != allChildren.end())
                ygWorld::world()->deleteTree(*iter);
            }
        sceneChildren.clear();
        parser.clear();
        loaded = false;
        loadFile(filename,preprocess);
        }
    else
        ygNode::message(msg);
    }


void ygScene::loadFile(const ygString& file,bool preprocess)
    {
    if (!loaded && filename.length() > 0)
        {
        if (first)
            {
            const vector<ygMessage*>& messages = getInitMessages();
            vector<ygMessage*>::const_iterator iter;
            for (iter=messages.begin(); iter != messages.end(); ++iter)
                {
                ygString msg; 
                (*iter)->writeString(msg);
                initMessages.push_back(msg);
                }
            first = false;
            }
        else
            {
            clearInitMessages();
            for (int i=0;i<initMessages.size();i++)
                addInitMessage(initMessages[i]);
            }
        int firstChild = numChildren();
        parser.parseFile(filename,this,preprocess);
        parser.replaceInitMessages(this);
        loaded = true;
        ygNode::reset();
        for (int i=firstChild;i<numChildren();i++)
            sceneChildren.insert(sceneChildren.begin(),child(i));
        }
    }


void ygScene::parseMessage(const ygMessage& msg,bool force)
    {
    ygMessage newMsg = msg;
    parser.replaceMessageName(&newMsg);
    bool updated;
    updated = parseNetMessage(newMsg,force);
    message(newMsg);
    if (updated)
        {
        //the node has been updated by a message
        eventOccurred("updated");
        }
    }

void ygScene::printMessages(ostream& s,int indent,bool comment)
    {
    if (comment)
        {
        const vector<ygString*>& comments = getComments();
        vector<ygString*>::const_iterator iter;
        for (iter=comments.begin(); iter != comments.end(); ++iter)
            {
            for (int i=0; i < indent; i++)
                s << "    ";
            s << "//" << *(*iter);
            }
        }
    int padIndent = 0;
    padIndent = indent*4+className().length()+name().length()+2;
    for (int i=0; i < indent; i++)
        s << "    ";
    s << className() << " " << name() << "(";
    const vector<ygMessage*>& messages = getNetMessages(false);
    for (int i=0;i < messages.size(); i++)
        {
        if (i > 0)
            {
            s << "," << endl;
            for (int j=0; j < padIndent; j++)
                s << " ";
            }
        s << *(messages[i]);
        }
    if (comment)
        {
        for (int i=0;i < initMessages.size(); i++)
            {
            if (i > 0)
                {
                s << "," << endl;
                for (int j=0; j < padIndent; j++)
                    s << " ";
                }
            s << initMessages[i];
            }
        }
    bool found=false;
    const map<ygString,vector<ygString> >& events = getEventMessages();
    map<ygString,vector<ygString> >::const_iterator i;
    for (i=events.begin(); i != events.end(); ++i)
        {
        const vector<ygString>& message = (*i).second;
        if (message.size() > 0)
            {
            if (found || messages.size() > 0)
                {
                s << "," << endl;
                for (int k=0; k < padIndent; k++)
                    s << " ";
                }
            found = true;
            for (int j=0;j < message.size(); j++)
                {
                if (j > 0)
                    {
                    s << "," << endl;
                    for (int k=0; k < padIndent; k++)
                        s << " ";
                    }
                s << "when(" << (*i).first << "," << message[j] << ")";
                }
            }
        }
    s << ")" << endl;
    }

