// Description: defines a space that is rectangular, spherical, cylindrical, point or infinite in volume
//
//<b>notes:</b>
//<ul>
//<li> derived nodes inherit all messages from their base classes
//<li> see reset method for default settings
//<li> this node can also be created with the alias: space
//<li> a volume message defines a box, sphere, cylinder, point or infinite volume as follows:
//<ul>
//<li> volume(<a href="box.gif">box</a> [X Y Z X Y Z]) - default points (-0.5 -0.5 -0.5) and (0.5 0.5 0.5) define the extent of the box
//<li> volume(sphere [X Y Z radius]) - default the point (0 0 0) represents the origin of a sphere 0.5 in radius
//<li> volume(cylinder [X Y Z X Y Z radius]) - default points (0 0 -0.5) and (0 0 0.5) are the center of top and bottom circles 0.5 in radius
//<li> volume(point [X Y Z]) - default point is (0 0 0)
//<li> volume(infinite) - an infinite volume
//</ul>
//</ul>
//
// Category: Foundation
// Author: Dave Pape
//           11/01/01
//           02/11/03 Dave Pape - corrected makeDebugOutline to properly remove any old debug geoset(s)
//           09/01/04 Alex Hill - changed to allow "volume" debug to be sent at any time
// Revision: 02/01/06 Alex Hill - broke volume message into volume type string and vector for optional extent
//
#include <Performer/pf/pfGroup.h>
#include <Performer/pf/pfGeode.h>
#include <Performer/pr/pfGeoSet.h>
#include <Performer/pfdu.h>
#include "ygWorld.h"
#include "ygUtil.h"
#include "ygNetMessages.h"
#include "ygBox.h"
#include "ygCylinder.h"
#include "ygSphere.h"
#include "ygPointVolume.h"
#include "ygInfiniteVolume.h"
#include "ygSpace.h"

using namespace std;

extern "C" ygNode* construct_ygSpace(const char* name,bool master) { return new ygSpace(name,master); }

#define SWAPFLOAT(a,b)  { float swap_temp; swap_temp=a; a=b; b=swap_temp; }

struct _ygSpacePrivateData
    {
    bool debugVolume;
    pfGeode * debugGeode;
    int netType;
    vector<float> netSize;
    };


ygSpace::ygSpace(const char* name,bool master) : ygNode(name,master)
    {
    setClassName("ygSpace",true);
    p_ = new struct _ygSpacePrivateData;
    volume_ = new ygInfiniteVolume;
    p_->debugGeode = NULL;
    //show a wireframe of the volume
    debugFlag("volume",getDebug("display"));
    p_->netType = YGVOL_INFINITE;
    //distribute the volume message
    addNetMessage("volume", &p_->netType, YG_NET_INT);
    addNetMessageMap("volume","infinite",YGVOL_INFINITE);
    addNetMessageMap("volume","box",YGVOL_BOX);
    addNetMessageMap("volume","sphere",YGVOL_SPHERE);
    addNetMessageMap("volume","cylinder",YGVOL_CYLINDER);
    addNetMessageMap("volume","point",YGVOL_POINT);
    addNetMessage("volume", &p_->netSize, YG_NET_VECTOR);
    }

ygSpace::~ygSpace(void)
    {
    if (volume_)
        delete volume_;
    }


void ygSpace::reset(void)
    {
    //set volume to infinite
    setVolume(new ygInfiniteVolume);
    resetNetMessage("volume");
    ygNode::reset();
    }


void ygSpace::message(const ygMessage& msg)
    {
    //set the shape and size of the volume enclosed
    if (msg == "volume")
        {
        if (p_->netType == YGVOL_BOX)
            {
            if (!createBoxVolume())
                msg.error(name(),"(wrong number of arguments - should be 'box [MINX MINY MINZ MAXX MAXY MAXZ]')");
            }
        else if (p_->netType == YGVOL_SPHERE)
            {
            if (!createSphereVolume())
                msg.error(name(),"(wrong number of arguments - should be 'sphere [CENTERX CENTERY CENTERZ RADIUS]')");
            }
        else if (p_->netType == YGVOL_CYLINDER)
            {
            if (!createCylinderVolume())
                msg.error(name(),"(wrong number of arguments - should be 'cylinder [BOTTOMX BOTTOMY BOTTOMZ TOPX TOPY TOPZ RADIUS]')");
            }
        else if (p_->netType == YGVOL_POINT)
            {
            if (!createPointVolume())
                msg.error(name(),"(wrong number of arguments - should be 'point [X Y Z]')");
            }
        else if (p_->netType == YGVOL_INFINITE)
            setVolume(new ygInfiniteVolume);
        else
            setVolume(NULL);
        }
    else
        ygNode::message(msg);
    }


bool ygSpace::createBoxVolume(void)
    {
    pfVec3 min,max;
    if (p_->netSize.size() == 0)
        {
        min.set(-0.5,-0.5,-0.5);
        max.set(0.5,0.5,0.5);
        }
    else if (p_->netSize.size() == 6)
        {
        min.set(p_->netSize[0],p_->netSize[1],p_->netSize[2]);
        max.set(p_->netSize[3],p_->netSize[4],p_->netSize[5]);
        if (min[0] > max[0])
            SWAPFLOAT(min[0],max[0]);
        if (min[1] > max[1])
            SWAPFLOAT(min[1],max[1]);
        if (min[2] > max[2])
            SWAPFLOAT(min[2],max[2]);
        }
    else
        return false;
    ygBox * box = new ygBox;
    box->min = min;
    box->max = max;
    setVolume(box);
    return true;
    }


bool ygSpace::createSphereVolume(void)
    {
    pfVec3 center;
    float radius;
    if (p_->netSize.size() == 0)
       {
       center.set(0,0,0);
       radius = 0.5;
       }
    else if (p_->netSize.size() == 4)
        {
        center.set(p_->netSize[0],p_->netSize[1],p_->netSize[2]);
        radius = p_->netSize[3];
        }
    else
        return false;
    ygSphere * sphere = new ygSphere;
    sphere->center = center;
    sphere->radius = radius;
    setVolume(sphere);
    return true;
    }


bool ygSpace::createCylinderVolume(void)
    {
    pfVec3 p0,p1,axis;
    float length,radius;
    if (p_->netSize.size() == 0)
        {
        p0.set(0,0,0.5);
        p1.set(0,0,-0.5);
        radius = 0.5;
        }
    else if (p_->netSize.size() == 7)
        {
        p0.set(p_->netSize[0],p_->netSize[1],p_->netSize[2]);
        p1.set(p_->netSize[3],p_->netSize[4],p_->netSize[5]);
        radius = p_->netSize[6];
        }
    else
        return false;
    ygCylinder *cylinder = new ygCylinder;
    cylinder->center = (p0+p1)/2.0f;
    cylinder->radius = radius;
    axis = p1 - p0;
    length = axis.length();
    if (length < 0.0000001f)
        axis.set(0,0,1);
    else
        axis /= length;
    cylinder->axis = axis;
    cylinder->halfLength = length/2.0f;
    setVolume(cylinder);
    return true;
    }


bool ygSpace::createPointVolume(void)
    {
    pfVec3 pos;
    if (p_->netSize.size() == 0)
        pos.set(0,0,0);
    else if (p_->netSize.size() == 3)
        pos.set(p_->netSize[0],p_->netSize[1],p_->netSize[2]);
    else
        return false;
    ygPointVolume * point = new ygPointVolume;
    *point = pos;
    setVolume(point);
    return true;
    }


void ygSpace::setVolume(ygVolume * vol)
    {
    if (volume_)
        delete volume_;
    if (vol)
        volume_ = vol;
    else
        volume_ = new ygInfiniteVolume;
    bool* debugDisplay = getDebug("display");
    if (*debugDisplay || debugNode())
        createDebugNode();
    }


void ygSpace::createDebugNode(void)
    {
    pfGeoSet * gset;
    pfGeode* geode = (pfGeode*)debugNode();
    if (geode)
        {
        while (geode->getNumGSets() > 0)
            {
            gset = geode->getGSet(0);
            if (gset)
                {
                geode->removeGSet(gset);
                pfDelete(gset);
                }
            }
        }
    else
        {
        geode = new pfGeode;
        if (!geode)
            {
            cerr << "ERROR: ygSpace::createDebugGeode: failed to create pfGeode\n";
            return;
            }
        setDebugNode((pfGroup*)geode);
        }
    gset = ygCreateVolumeGeoset(volume_,YG_STYLE_WIREFRAME,debugColor());
    if (gset)
        geode->addGSet(gset);
    else
        createCrossGeosets(geode);
    }


void ygSpace::createCrossGeosets(pfGeode* geode)
    {
    pfMatrix matrix;
    pfGeoSet* gset;
    //create X coordinate axis
    gset = makePlaneGeoset();
    pfGeoState* gstate = new pfGeoState;
    gstate->setMode(PFSTATE_ENWIREFRAME,PF_ON);
    gset->setGState(gstate);
    gset->setAttr(PFGS_COLOR4,PFGS_OVERALL,debugColor(),NULL);
    matrix.makeScale(0.5,0.01,1);
    matrix.postRot(matrix,90.0,1,0,0);
    pfdXformGSet(gset,matrix);
    geode->addGSet(gset);    
    gset = makePlaneGeoset();
    gset->setGState(gstate);
    gset->setAttr(PFGS_COLOR4,PFGS_OVERALL,debugColor(),NULL);
    matrix.makeScale(0.5,0.01,1);
    pfdXformGSet(gset,matrix);
    geode->addGSet(gset);  
    //create Y coordinate axis  
    gset = makePlaneGeoset();
    gset->setGState(gstate);
    gset->setAttr(PFGS_COLOR4,PFGS_OVERALL,debugColor(),NULL);
    matrix.makeScale(0.01,0.5,1);
    matrix.postRot(matrix,90.0,0,1,0);
    pfdXformGSet(gset,matrix);
    geode->addGSet(gset);    
    gset = makePlaneGeoset();
    gset->setGState(gstate);
    gset->setAttr(PFGS_COLOR4,PFGS_OVERALL,debugColor(),NULL);
    matrix.makeScale(0.01,0.5,1);
    pfdXformGSet(gset,matrix);
    geode->addGSet(gset);  
    //create Z coordinate axis  
    gset = makePlaneGeoset();
    gset->setGState(gstate);
    gset->setAttr(PFGS_COLOR4,PFGS_OVERALL,debugColor(),NULL);
    matrix.makeScale(1,0.01,0.5);
    matrix.postRot(matrix,90.0,0,1,0);
    pfdXformGSet(gset,matrix);
    geode->addGSet(gset);    
    gset = makePlaneGeoset();
    gset->setGState(gstate);
    gset->setAttr(PFGS_COLOR4,PFGS_OVERALL,debugColor(),NULL);
    matrix.makeScale(0.01,1,0.5);
    matrix.postRot(matrix,90.0,1,0,0);
    pfdXformGSet(gset,matrix);
    geode->addGSet(gset);
    }

 
pfGeoSet* ygSpace::makePlaneGeoset(void)
    {
    pfGeoSet *gset = new pfGeoSet;
    int *lengths = (int *) pfMalloc(1 * sizeof(int),pfGetSharedArena());
    gset->setPrimType(PFGS_POLYS);
    gset->setNumPrims(1);
    lengths[0] = 4;
    gset->setPrimLengths(lengths);
    pfVec3 * verts = (pfVec3 *) pfMalloc(4 * sizeof(pfVec3),pfGetSharedArena());
    verts[0][0] = -1; verts[0][1] = -1; verts[0][2] = 0;
    verts[1][0] = 1; verts[1][1] = -1; verts[1][2] = 0;
    verts[2][0] = 1; verts[2][1] = 1; verts[2][2] = 0;
    verts[3][0] = -1; verts[3][1] = 1; verts[3][2] = 0;
    gset->setAttr(PFGS_COORD3,PFGS_PER_VERTEX,verts,NULL);
    gset->setLineWidth(1.0);
    gset->setDrawMode(PFGS_WIREFRAME,1);
    return gset;
    }
