// Description: aggregates the user related functions of wands, navigators, and trackers
//
//<b>notes:</b>
//<ul>
//<li> derived nodes inherit all messages from their base classes
//<li> see reset method for default settings
//<li> this node can also be created with the alias: user
//<li> all messages are passed on to the associated navigator
//<li> this node should a parent of any <a href="ygCAVEHead.html">ygCAVEHead</a> or <a href="ygCAVENavigator.html">ygCAVENavigator</a> nodes
//<li> this node is required for userTrigger and wandTrigger detection
//</ul>
//
// Category: User
// Author: Dave Pape
//           11/01/01
//           12/20/02 Alex Hill - added ygNode message filtering
// Revision: 06/28/05 Alex Hill - added optional CAVE sensor and navigation updating
//
#include <vector>
#include <Performer/pf/pfGroup.h>
#include <pfcave.h>
#include "ygNodeDB.h"
#include "ygWorld.h"
#include "ygNetMessages.h"
#include "ygNavigator.h"
#include "ygUser.h"

#define NULL_SENSOR "null_ycdq"

using namespace std;

extern "C" ygNode* construct_ygUser(const char* name,bool master) { return new ygUser(name,master); }

struct _ygUserPrivateData
    {
    ygNavigator * navigator;
    ygNode* head;
    vector<ygNode*> wands;
    vector<ygNode*> sensors;
    ygString netHead, netWands, netSensors;
    ygString filterMessages;
    bool caveTracking;
    };


ygUser::ygUser(const char* name,bool master) : ygTransform(name,master)
    {
    ygString str;
    setClassName("ygUser",true);
    p_ = new struct _ygUserPrivateData;
    p_->head = NULL;
    p_->navigator = NULL;
    p_->caveTracking = false;
    //distribute the name of the head node
    addNetMessage("head", &p_->netHead, YG_NET_STRING);
    //dont include in documentation
    dontSaveNetMessage("head");
    dontSendNetMessage("head");
    //distribute the names of all wand nodes
    addNetMessage("wand", &p_->netWands, YG_NET_STRING);
    //dont include in documentation
    dontSaveNetMessage("wand");
    dontSendNetMessage("wand");
    //distribute the names of all sensor nodes
    addNetMessage("sensor", &p_->netSensors, YG_NET_STRING);
    //dont include in documentation
    dontSaveNetMessage("sensor");
    ygWorld::world()->addUser(this);
    p_->filterMessages = " when addChild removeChild debug print reset event unreliableMessage reliableMessage resetTree position orientation size updateInterval ";
    }


ygUser::~ygUser(void)
    {
    ygWorld::world()->removeUser(this);
    delete p_;
    }


void ygUser::message(const ygMessage& msg)
    {
    ygString paddedMessage = " ";
    paddedMessage += msg.message();
    paddedMessage += " ";
    //hide all avatar geometry at the local site (deprectated)
    if (msg == "hideLocal")
        {
        if (msg.numArgs() > 0)
            hideLocal(msg.boolArg(0));
        else
            hideLocal();
        }
    else if (msg == "head")
        {
        p_->head = ygNodeDB::find(p_->netHead);
        storeHeadKey();
        }
    else if (msg == "wand")
        {
        ygString *wandName;
        int i=0;
        p_->wands.clear();
        while (wandName = p_->netWands.nextToken(" ",&i))
            {
            ygNode * n = ygNodeDB::find(*wandName);
            if (n)
                p_->wands.push_back(n);
            delete wandName;
            }
        storeWandKey();
        }
    else if (msg == "sensor")
        {
        ygString *sensorName;
        int i=0;
        p_->sensors.clear();
        while (sensorName = p_->netSensors.nextToken(" ",&i))
            {
            ygNode* n;
            if (*sensorName == NULL_SENSOR)
                n = NULL;
            else
                n = ygNodeDB::find(*sensorName);
            if (n)
                p_->sensors.push_back(n);
            delete sensorName;
            }
        }
    //pass all other messages on to the associated navigator
    else if (navigator() && !p_->filterMessages.contains(paddedMessage))
        {
        navigator()->parseMessage(msg);
        }
    else
        ygTransform::message(msg);
    }


void ygUser::hideLocal(bool hide)
    {
    if (hide)
        pfnode()->setTravMask(PFTRAV_DRAW, 0, PFTRAV_SELF, PF_SET);
    else
        pfnode()->setTravMask(PFTRAV_DRAW, 0xffffffff, PFTRAV_SELF, PF_SET);
    }


void ygUser::app(void)
    {
    if (p_->caveTracking && !CAVEDistribMaster())
        {
        //set CAVE navigation matrix
        pfMatrix mat;
        CAVENavGetMatrix(mat.mat);
        mat.invertOrtho(mat);
        pfCoord coord;
        mat.getOrthoCoord(&coord);
        setPosition(coord.xyz);
        setOrientation(coord.hpr[1],coord.hpr[2],coord.hpr[0]);
        //set CAVE head position
        pfVec3 pos, ori;
        for (int i=0;i<numSensors();i++)
            {
            if (sensor(i))
                {
                //get the CAVE sensor position
                CAVEGetSensorPosition(CAVESENSOR(i),CAVE_TRACKER_FRAME,pos.vec);
                //get the CAVE sensor orientation
                CAVEGetSensorOrientation(CAVESENSOR(i),CAVE_TRACKER_FRAME,ori.vec);
                ((ygTransform*)sensor(i))->setPosition(pos);
                ((ygTransform*)sensor(i))->setOrientation(ori);
                }
            }
        }
    ygTransform::app();
    }


void ygUser::setCAVETracking(bool use)
    {
    p_->caveTracking = use;
    }


bool ygUser::isCAVETracking(void)
    {
    return p_->caveTracking;
    }


void ygUser::setHead(ygNode* head)
    {
    p_->head = head;
    storeHeadKey();
    }


void ygUser::addWand(ygNode* wand)
    {
    int i;
    for (i=0; i < p_->wands.size(); i++)
        if (p_->wands[i] == wand)
            return;
    p_->wands.push_back(wand);
    storeWandKey();
    }


void ygUser::addSensor(ygNode* sensor,int num)
    {
    int i;
    while (numSensors() <= num)
        p_->sensors.push_back(NULL);
    p_->sensors[num] = sensor;
    storeSensorKey();
    }


void ygUser::setNavigator(ygNavigator *nav)
    {
    p_->navigator = nav;
    }


ygNavigator * ygUser::navigator(void) const
    {
    return p_->navigator;
    }


ygNode * ygUser::head(void) const
    {
    return p_->head;
    }


int ygUser::numWands(void) const
    {
    return p_->wands.size();
    }


ygNode * ygUser::wand(int i) const
    {
    if ((0 <= i) && (i < numWands()))
        return p_->wands[i];
    else
        return NULL;
    }


int ygUser::numSensors(void) const
    {
    return p_->sensors.size();
    }


ygNode * ygUser::sensor(int i) const
    {
    if ((0 <= i) && (i < numSensors()))
        return p_->sensors[i];
    else
        return NULL;
    }


ygUser* ygUser::findUserAncestor(ygNode * node)
    {
    ygUser * user = NULL;
    while ((node) && (!user))
        {
        if (node->isOfClass("ygUser"))
            user = (ygUser*) node;
        else
            node = node->parent();
        }
    return user;
    }


/***  Net functions  ***/

void ygUser::storeWandKey(void)
    {
    int i;
    p_->netWands.clear();
    for (i=0; i < numWands(); i++)
        {
        p_->netWands += wand(i)->name();
        p_->netWands += " ";
        }
    netMessageChanged("wand");
    }


void ygUser::storeSensorKey(void)
    {
    int i;
    p_->netSensors.clear();
    for (i=0; i < numSensors(); i++)
        {
        if (sensor(i))
            p_->netSensors += sensor(i)->name();
        else
            p_->netSensors += NULL_SENSOR;
        p_->netSensors += " ";
        }
    netMessageChanged("sensor");
    }


void ygUser::storeHeadKey(void)
    {
    if (head())
        {
        p_->netHead = head()->name();
        netMessageChanged("head");
        }
    }
